#!/usr/bin/php5
<?

/**

  Der Listener schreibt eine oder mehrere Aenderungen in eine oder mehrere Dateien, Dateiname ist jeweils der Unix-Timestamp.
  Anschliessend wird dieses PHP-Script aufgerufen und der Pfad zur Konfiguration sowie die jeweilige Schul-ID uebergeben.

  Grobe Vorgehensweise:
  - Lesen der Konfiguration
  - Erstellen der Dateiliste
  - Login bei der jeweiligen WWS
  - Lesen der Datei und Abarbeiten jedes Zeilen-Trippels
    - Synchronisation des betreffenden Kontos
      - Angelegen
      - Aendern
      - Zum Loeschen markieren
      - Bei Gruppen: Mitgliederliste aus UCS zwischenspeichern
  - Abarbeiten der Mitgliederlisten
  - Loeschen von Konten

  Anmerkungen:
  - Die Synchronisation der Konten erfolgt ueber die UUID.
  - Die Mitgliederliste einer Gruppe wird im Attribut "uniqueMember" geloggt ("uid=test.schueler,cn=schueler,cn=users,ou=Testschule,dc=ucs-testschule,dc=intranet"). Hier erfolgt die Synchronisation ueber den Login/Alias.
  - Eine Aenderung des Nutzernamens in UCS entspricht wird vom Listener als Loeschen/Neuanlegen geloggt. Daher das Loeschen zum Schluss.
  - Beim CSV-Import loggt der Listener die Gruppenliste nach Anlegen des ersten Nutzers. Daher kann sie erst zum Schluss abgearbeitet werden, wenn auch alle Nutzer angelegt wurden.

*/

  echo sprintf("WWS-Sync-Script [%s] started.\n",$argv[2]);

  // Log-File
  $path_log='/var/log/univention-digionline-webweaver/sync.log';

  // JSON-Datei mit Konfiguration
  $config_filename=$argv[1];

  // Name der Schule (Filter fuer Konfiguration)
  $sync_school=$argv[2];

  // Einstellungen
  $sync_data=array();
  /**
    [sync_school] => TtenggelSchule,                                                                      // Name der Schule, nur a-z0-9_-
    [ServerURL] => www.wwschool.de, [ServerLogin] => sync@wwschool.de, [ServerPasswd] => *******,         // Zugangsdaten
    [lehrer] => 0, [schueler] => 1, [mitarbeiter] => 2, [l_und_m] => 3, [gruppen] => 18, [klassen] => 19, // Mapping der Rollen
    [error_email] => sync@wwschool.de,                                                                    // E-Mail-Adresse fuer Fehlermeldungen
    [ExportDirectory] => /var/spool/univention-digionline-webweaver/TtenggelSchule                        // Verzeichnis mit Export-Dateien
  */
  if(!$sync_array=json_decode(file_get_contents($config_filename),true)) {
    write_log('Die Konfigurationsdatei ist nicht lesbar.','FATAL');
  }
  foreach($sync_array as $nr=>$data) {
    if($data['sync_school']==$sync_school) {
      $sync_data=$sync_array[$nr];
    }
  }
  if(!$sync_data) {
    write_log(sprintf('Die Konfigurationsdatei enthaelt keine Eintrage fuer "%s".',$sync_school),'FATAL');
  }

  // Administrierbare Rollen (Nutzer und Gruppen/Klasse)
  $wws_types=array();

  // Login der Institution
  $root_login='';

  // Domain der Institution
  $root_domain='';

  // Session-ID
  $session_id='';

  // Queue mit eventuell zu loeschenden Konten: $uuid=>$ident
  $queue_delete=array();

  // Queue mit Mitgliederlisten: $uuid=>array()
  $queue_memberships=array();

  // Alle vom Listener generierten Dateien lesen
  $filelist=read_export($sync_data['ExportDirectory']);

  if(count($filelist)) {
    $soap_client=new SoapClient('https://'.$sync_data['ServerURL'].'/soap.php?wsdl',array('encoding'=>'UTF-8','trace'=>1,'features'=>SOAP_SINGLE_ELEMENT_ARRAYS,'exceptions'=>0));
    // Login
    $result=(array)$soap_client->request(array(array('method'=>'login','login'=>$sync_data['ServerLogin'],'password'=>$sync_data['ServerPasswd'])));
    if(('OK'===$result[0]['return']) and ($session_id=$result[1]['session_id'])) {
      $root_domain=preg_replace('/^.*?@/','',$result[0]['user']['login']);
      $wws_types=array_intersect($result[1]['admin_types'],array(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,18,19)); // Nutzer-Rollen und Institutions-Gruppen/-Klassen
      $arr_memberships=$result[0]['member'];
      if($root_domain and $wws_types and $arr_memberships) {
        $root_login='info@'.$root_domain;
        $is_admin=false;
        // Test auf Admin-Rechte in der Institution
        foreach($arr_memberships as $t_membership) {
          if($t_membership['login']===$root_login) {
            $is_admin=((false!==array_search('admin_admin',$t_membership['base_rights'])) and (false!==array_search('admin_admin',$t_membership['member_rights'])));
          }
        }
        if($is_admin) {
          // Die Konten der Schule mit den Export-Dateien synchronisieren
          sync_school();
        } else {
          write_log('Das Synchronisationskonto benoetigt volle Admin-Rechte in der Institution.','FATAL');
        }
      } else {
        write_log('Unerwartete Antwort bei der SOAP-Kommunikation.','FATAL',$result);
      }
      $soap_client->request(array(array('method'=>'set_session','session_id'=>$session_id),array('method'=>'logout')));
    } else {
      write_log('Login verweigert.','FATAL');
    }
  }

  echo sprintf("WWS-Sync-Script [%s] ended.\n",$argv[2]);

  /**
   * E-Mail ueber SOAP verschicken, falls eingeloggt. Keine Fehlerbehandlung
   * @param string $to
   * @param string $subject
   * @param string $text
   */
  function send_mail($to,$subject,$text) {
    global $soap_client,$session_id;
    if($to and strpos($to,'@') and $subject and $soap_client and $session_id) {
      // Aufruf direkt, damit es nicht zu einer Endlosschleife kommen kann.
      $soap_client->request(array(
        array('method'=>'set_session','session_id'=>$session_id),
        array('method'=>'set_focus','object'=>'mailbox'),
        array('method'=>'send_mail','to'=>$to,'subject'=>$subject,'text'=>$text),
      ));
    }
  }

  /**
   * Eingeloggt: SOAP-Request schicken
   * @param string $selector Selektor fuer $ident
   * @param string $ident Login, UUID usw. fuer User-Objekt
   * @param bool $access_denied_is_fatal Fehler-Level bei set_focus()
   * @param string $object Objekt, auf das der Focus gesetzt werden soll
   * @param string $arr_request Der Request
   */
  function soap_request($selector,$ident,$object,$access_denied_is_fatal,$arr_request) {
    global $session_id,$soap_client;
    $arr_result=false;
    $arr_complete_request=array(
      array('method'=>'set_session','session_id'=>$session_id),
      array('method'=>'set_focus',$selector=>$ident,'object'=>$object),
      $arr_request,
    );
    $i=0;
    $arr_response=NULL;
    while(!is_array($arr_response) and ($i<5)) {
      if($i>0) {
        write_log('Fehler bei der SOAP-Kommunikation','LOG',$arr_response);
        sleep(1);
      }
      $arr_response=$soap_client->request($arr_complete_request);
      $i++;
    }
    if(!is_array($arr_response)) {
      write_log('Fehler bei der SOAP-Kommunikation','FATAL',$arr_response);
    } elseif(($arr_response[0]['return']!=='OK')) { // set_session
      write_log('Sitzung abgelaufen.','FATAL');
    } elseif($arr_response[1]['return']!=='OK') { // set_focus
      write_log(sprintf('Zugriff auf Objekt "%s" verweigert oder Objekt nicht gefunden.',$ident),$access_denied_is_fatal?'FATAL':'ERROR');
    } else {
      $arr_result=$arr_response[2]; // Antwort auf $arr_request
    }
    return $arr_result;
  }

  /**
   * SOAP-Request auf Institutions-Administration
   * @param array $arr_request
   * @return array
   */
  function soap_request_root_admin($arr_request) {
    global $root_login;
    return soap_request('login',$root_login,'administration',true,$arr_request);
  }

  /**
   * SOAP-Request auf Gruppen/Klassen-Administration
   * @param string $group_uuid
   * @param array $arr_request
   * @return array
   */
  function soap_request_group_admin($group_uuid,$arr_request) {
    return soap_request('uid',$group_uuid,'administration',false,$arr_request);
  }

  /**
   * User-Liste der Gruppe/Klasse einlesen
   * @param string $uuid
   */
  function soap_request_get_members($uuid) {
    $arr_request=array('method'=>'get_users');
    $arr_users=array();
    $ready=false;
    while(!$ready) {
      $ready=true;
      $arr=soap_request('uid',$uuid,'administration',false,$arr_request);
      if(!in_array($arr['return'],array('OK','RESUME','RETRY'))) {
        write_log(sprintf('Fehler beim Einlesen der Nutzerliste von "%s".',$ident),'ERROR');
      } else {
        if(isset($arr['users'])) {
          foreach($arr['users'] as $t_user) {
            $arr_users[]=$t_user;
          }
        }
        if($arr['return']!=='OK') {
          $ready=false;
          if($t_offset=@$arr['offset']) {
            $arr_request['offset']=$t_offset;
          }
        }
      }
    }
    return $arr_users;
  }

  /**
   * Sortiertes Array mit Pfaden zu allen numerischen Dateien in $main_dir
   * @param string $main_dir
   * @return array string
   */
  function read_export($main_dir) {
    $filelist=array();
    foreach(scandir($main_dir) as $filename) {
      if(preg_match('@^[0-9]{2,}$@',$filename)) {
        $filelist[$filename]=$main_dir.'/'.$filename;
      }
    }
    ksort($filelist);
    return $filelist;
  }

  /**
   * Schule synchronisieren
   * @return void
   */
  function sync_school() {
    global $sync_school,$sync_data,$filelist,$wws_types,$queue_memberships,$queue_delete;
    // Abarbeiten der Synchronisationsdateien
    foreach($filelist as $file) {
      $s=file_get_contents($file);
      $lines=preg_split("/\n/",$s,-1,PREG_SPLIT_NO_EMPTY);
      $n=count($lines);
      if(!$n or ($n%3)) { // Anzahl der Zeilen nicht durch drei teilbar
        write_log(sprintf('Datei "%s" fehlerhaft. Anzahl Zeilen: %s',$file,(string)$n),'FATAL',$s);
      }
      // Immer drei Zeilen gehoeren zusammen: Objekt, Neu, Alt
      for($i=0;$i<$n;$i=$i+3) {
        // Zerlegen der 1. Zeile des Trippels "cn=Testschule-Testklasse,cn=klassen,cn=schueler,cn=groups,ou=Testschule,dc=ucs-testschule,dc=intranet"
        $dn=array();
        foreach(explode(',',trim($lines[$i],'"')) as $item) {
          list($var,$val)=explode('=',$item);
          $dn[$var][]=$val;
        }
        if(@$dn['ou'][0]===$sync_school) { // Gehoert der Datensatz zur zu synchronisierenden Schule?
          if($cn=@$dn['cn']) {
            // Was ist anhand der Sync-Zeilen zu tun?
            $action='';
            // 2. Zeile des Trippels: Neuer Datensatz als JSON
            $record_new=json_decode($lines[$i+1],true);
            // 3. Zeile des Trippels: Neuer Datensatz als JSON
            $record_old=json_decode($lines[$i+2],true);
            if(!is_array($record_new) or !is_array($record_old)) {
              write_log('Fehler in Sync-Datei','FATAL',$lines[$i]."\n".$lines[$i+1]."\n".$lines[$i+2]);
            }
            if($record_new and $record_old) {
              $action='edit';
            } elseif(!$record_new and $record_old) {
              $action='del';
            } elseif($record_new and !$record_old) {
              $action='add';
            } else {
              $action='error';
            }
            // UUIDs sind Primaerschluessel, "UCS-" wird vorangestellt
            $uid_n=@$record_new['entryUUID']?('UCS-'.$record_new['entryUUID']):NULL;
            $uid_a=@$record_old['entryUUID']?('UCS-'.$record_old['entryUUID']):NULL;
            $wws_type=-1;
            $ident=NULL;
            // Rolle und Ident bestimmen
            if(in_array('groups',$cn) and in_array('klassen',$cn) and in_array('schueler',$cn)) { // Klasse
              $wws_type=$sync_data['klassen'];
              $ident=preg_replace('@^'.preg_quote($sync_school.'-').'@','',$cn[0]);
            } elseif(in_array('groups', $cn) and in_array('schueler', $cn)) { // Gruppe
              $wws_type=$sync_data['gruppen'];
              $ident=preg_replace('@^'.preg_quote($sync_school.'-').'@','',$cn[0]);
            } elseif(in_array('mitarbeiter', $cn) and in_array('users', $cn)) { // Mitarbeiter
              $wws_type=$sync_data['mitarbeiter'];
              $ident=$dn['uid'][0];
            } elseif(in_array('lehrer und mitarbeiter', $cn) and in_array('users', $cn)) { // Lehrer und Mitarbeiter
              $wws_type=$sync_data['l_und_m'];
              $ident=$dn['uid'][0];
            } elseif(in_array('schueler', $cn) and in_array('users', $cn)) { // Schueler
              $wws_type=$sync_data['schueler'];
              $ident=$dn['uid'][0];
            } elseif(in_array('lehrer', $cn) and in_array('users', $cn)) { // Lehrer
              $wws_type=$sync_data['lehrer'];
              $ident=$dn['uid'][0];
            }
            if($wws_type>=0) {
              while(($wws_type<16) and ($wws_type>0) and !in_array($wws_type,$wws_types)) {
                // Falls nicht alle Nutzer-Rollen freigeschaltet sind, die naechst niedrigere waehlen
                $wws_type--;
              }
              if(in_array($wws_type,$wws_types)) {
                if($uid_n) {
                  // Eventuellen Eintrag aus Loesch-Liste entfernen: Konto wird nur umbenannt
                  unset($queue_delete[$uid_n]);
                }
                if($action==='del') {
                  // In Loesch-Liste eintragen: Konto wird eventuell geloescht
                  $queue_delete[$uid_a]=$ident;
                } elseif(($action==='add') or ($action==='edit')) {
                  sync_user($ident,$uid_n,$record_new,$record_old,$wws_type);
                }
              } else {
                write_log(sprintf('Konfigurationsfehler: Die Rolle "%s" von "%s" kann nicht gemappt werden',$wws_type,$ident),'ERROR');
              }

            }
          }
        }
      }
    }
    // Mitgliedschaften
    foreach(array_keys($queue_memberships) as $uuid) {
      sync_memberships($uuid);
    }
    // Zum Schluss: Loesch-Liste abarbeiten
    foreach($queue_delete as $uuid=>$ident) {
      $response=soap_request_root_admin(array('method'=>'delete_user','uid'=>$uuid));
      if($response['return']==='OK') {
        write_log(sprintf('Konto "%s" ("%s") geloescht.',$uuid,$ident),'LOG');
      } elseif($response['errno']==154) {
        write_log(sprintf('Konto "%s" ("%s") wurde nicht geloescht, da es nicht existiert.',$uuid,$ident),'LOG');
      } else {
        write_log(sprintf('Konto "%s" ("%s") konnte nicht geloescht werden.',$uuid,$ident),'ERROR',$response);
      }
    }
    foreach($filelist as $file) {
      unlink($file); // Abgearbeitete Export-Dateien loeschen
    }

  }

  /**
   * Konto Synchronisieren
   * @param string $name Konto-Name in UCS
   * @param string $uuid UUID
   * @param array $record_new
   * @param array $record_old
   * @param int $type Rolle
   * @return void
   */
  function sync_user($name,$uuid,$record_new,$record_old,$type) {
    global $root_domain,$queue_memberships;
    $arr_properties=array();
    $is_group=($type>=16); // Synchronisieren wir eine Gruppe/Klasse?
    // Geaendertes Passwort?
    if(strlen(@$record_new['userPassword']) and ($record_new['userPassword']!==@$record_old['userPassword'])) {
      $new_crypt=$record_new['userPassword']; // SHA-256 incl. {crypt}
    } elseif(strlen(@$record_new['sambaNTPassword']) and ($record_new['sambaNTPassword']!==@$record_old['sambaNTPassword'])) {
      $new_crypt='{ntlm}'.$record_new['sambaNTPassword']; // NTLM
    } else {
      $new_crypt=NULL;
    }

    // Username und Alias
    $username=make_username($name);
    $login=$username.'@'.$root_domain;
    if(!$is_group and ($username!==strtolower($name))) {
      // Login mit Usernamen (Sonderzeichen) von UCS ermoeglichen
      $arr_properties['alias']=$name.'@'.$root_domain;
    } else {
      $arr_properties['alias']='';
    }

    // Fullname
    if($is_group) {
      $arr_properties['fullname']=$name;
    } elseif(strlen(@$record_new['cn'])) {
      $arr_properties['fullname']=$record_new['cn']; // $record_new['displayName'];
    }
    // Beschreibung
    if($is_group and strlen(@$record_new['description'])) {
      $arr_properties['description']=$record_new['description'];
    }
    // E-Mail
    if(!$is_group and strlen(@$record_new['mail'])) {
      $arr_properties['authentication_emailaddress']=$record_new['mail'];
    }

    $response=soap_request_root_admin(array('method'=>'get_user','uid'=>$uuid));
    if($response['return']!=='OK') { // Konto existiert nicht
      $response=soap_request_root_admin(array('method'=>'add_user','uid'=>$uuid,'type'=>$type,'user_name'=>$username));
      if($response['return']!=='OK') {
        write_log(sprintf('Anlegen des Kontos "%s", Username "%s" fehlgeschlagen.',$uuid,$username),'ERROR',$response);
        return false;
      }
      write_log(sprintf('Konto "%s" ("%s") angelegt.',$uuid,$username),'LOG');
    }
    $wws_user=$response['user'];

    // Setzen aller geaenderten Eigenschaften
    $arr_update=array();
    if($login!==$wws_user['login']) {
      $arr_update['user_name']=$username;
    }
    if(!is_null($new_crypt)) {
      $arr_update['crypt']=$new_crypt;
    }
    foreach($arr_properties as $t_var=>$t_val) {
      if((string)$t_val!==(string)@$wws_user[$t_var]) {
        $arr_update[$t_var]=(string)$t_val;
      }
    }
    if($arr_update) {
      $arr_update['revoke_base_right']='password'; // Aenderung das Passworts verbieten
      $response=soap_request_root_admin(array('method'=>'set_user','uid'=>$uuid)+$arr_update);
      if($response['return']!=='OK') {
        write_log(sprintf('Nicht alle Eigenschaften des Kontos "%s", Username "%s" konnten gesetzt werden.',$uuid,$username),'ERROR',$response);
      }
      $wws_user=$arr_update+$wws_user;
    }

    // Mitglieder einer Gruppe
    if($is_group and (isset($record_new['uniqueMember']) or isset($record_old['uniqueMember']))) {
      $queue_memberships[$uuid]=array(
        'user'   => $wws_user,
        'record' => $record_new,
      );
    }
  }

 /**
  * Mitgliedschaften synchronisieren
  * @param string $uuid
  */
 function sync_memberships($uuid) {
    global $root_domain,$sync_school,$queue_memberships;
    $wws_user=$queue_memberships[$uuid]['user'];
    $record=$queue_memberships[$uuid]['record'];
    // Erster Lehrer in $record_new
    $login_first_teacher=NULL;
    // Mitgliederliste aus WWS lesen
    $wws_members=array();
    foreach(soap_request_get_members($uuid) as $t_user) {
      if(substr($t_user['uid'],0,4)==='UCS-') { // Nur Mitglieder, die aus UCS kommen
        $wws_members[]=$t_user['login'];
      }
    }
    // Uebergebene Mitgliederliste parsen
    $ucs_members=array();
    if(@$record['uniqueMember']) {
      /*
            [0] => uid=test.lehrer,cn=lehrer,cn=users,ou=Testschule,dc=ucs-testschule,dc=intranet
            [1] => uid=test.schueler,cn=schueler,cn=users,ou=Testschule,dc=ucs-testschule,dc=intranet
            [2] => uid=dsfgdsfgdfsg,cn=mitarbeiter,cn=users,ou=Testschule,dc=ucs-testschule,dc=intranet
            [3] => uid=Lehrer-und-Mitarbeiter,cn=lehrer und mitarbeiter,cn=users,ou=Testschule,dc=ucs-testschule,dc=intranet
      */
      foreach((array)$record['uniqueMember'] as $t_dn) {
        if(preg_match('@,ou='.preg_quote($sync_school,'@').',@', $t_dn)) { // Nutzer in dieser Schule?
          if($t_name=preg_filter('@^uid=([^,]+).*@','$1',$t_dn)) { // UCS-Usernamen filtern
            $t_login=make_username($t_name).'@'.$root_domain; // WWS-Login
            $ucs_members[]=$t_login;
            if(preg_match('@,cn=lehrer,@',$t_dn)) { // Ist das ein Lehrer?
              if(is_null($login_first_teacher)) {
                $login_first_teacher=$t_login; // Erster Lehrer in diesem Record
              }
            }
          }
        }
      }
    }
    // Mitglieder entfernen
    foreach(array_diff($wws_members,$ucs_members) as $t_login) {
      $response=soap_request_group_admin($uuid,array('method'=>'leave_user','login'=>$t_login));
      if($response['return']!=='OK') {
        write_log(sprintf('Mitglied "%s" kann nicht aus Gruppe/Klasse "%s" entfernt werden.',$t_login,$wws_user['login']),'ERROR',$response);
      }
      if($wws_user['moderator']['login']===$t_login) {
        soap_request_root_admin(array('method'=>'set_user','uid'=>$uuid,'moderator'=>''));
        $wws_user['moderator']=NULL;
      }
    }
    // Mitglieder aufnehmen
    foreach(array_diff($ucs_members,$wws_members) as $t_login) {
      $response=soap_request_group_admin($uuid,array('method'=>'join_user','login'=>$t_login));
      if($response['return']!=='OK') {
        write_log(sprintf('Nutzer "%s" kann nicht in die Gruppe/Klasse "%s" aufgenommen werden.',$t_login,$wws_user['login']),'ERROR',$response);
      }
    }
    // Moderator setzen
    if(($wws_user['type']==19) and $login_first_teacher and !@$wws_user['moderator']['login']) {
      soap_request_root_admin(array('method'=>'set_user','uid'=>$uuid,'moderator'=>$login_first_teacher));
    }
  }

  /**
   * Ins Logfile schreiben, ggf. Meldung per Mail verschicken, ggf. Ausfuehrung abbrechen
   * @param string $message
   * @param string $severity
   * @param string $debuginfo
   */
  function write_log($message,$severity,$debuginfo=NULL) {
    global $path_log,$sync_data;
    $severity=strtoupper($severity);
    $addr=@$sync_data['error_email'];
    if($debuginfo and !is_scalar($debuginfo)) {
      $debuginfo=print_r($debuginfo,true);
    }
    if($debuginfo) {
      $debuginfo="\n\n\nDebug-Informationen:\n".$debuginfo;
    }
    switch($severity) {
      case 'DEBUG':
        break;
      case 'LOG':
        // send_mail($addr,'Meldung der Synchronisation',"Die Synchronisation meldet:\n" . $message.$debuginfo);
        break;
      case 'ERROR':
        send_mail($addr,'Fehler bei der Synchronisation',"Bei der Synchronisation ist ein Fehler aufgetreten.\n\nMeldung:\n" . $message.$debuginfo);
        break;
      case 'FATAL':
        send_mail($addr,'Schwerwiegender Fehler bei der Synchronisation',"Bei der Synchronisation ist ein schwerer Fehler aufgetreten, der eine Ueberpruefung und Resynchronisation erfordert.\n\nMeldung:\n" . $message.$debuginfo);
        break;
    }
    if($path_log) {
      error_log(date('Y-m-d H-i-s')."\t".$severity."\t".getmypid()."\t".preg_replace('@\\s+@',' ',$message)."\n",3,$path_log);
    }
    if($severity==='FATAL') {
      trigger_error($message,E_USER_ERROR);
    }
  }

  /**
   * Fuer WWS gueltigen User-Namen erstellen
   * @param string $s Wunschvorstellung
   * @return string
   */
  function make_username($s) {
    $s=strtr($s,array(
        '@'=>'-at-','Š'=>'S','Ž'=>'Z','š'=>'s','ž'=>'z','Ÿ'=>'Y','À'=>'A','Á'=>'A','Â'=>'A','Ã'=>'A','Å'=>'A','Ç'=>'C','È'=>'E','É'=>'E','Ê'=>'E','Ë'=>'E','Ì'=>'I','Í'=>'I',
        'Î'=>'I','Ï'=>'I','Ñ'=>'N','Ò'=>'O','Ó'=>'O','Ô'=>'O','Õ'=>'O','Ù'=>'U','Ú'=>'U','Û'=>'U','Ý'=>'Y','à'=>'a','á'=>'a','â'=>'a','ã'=>'a','å'=>'a','ç'=>'c','è'=>'e',
        'é'=>'e','ê'=>'e','ë'=>'e','ì'=>'i','í'=>'i','î'=>'i','ï'=>'i','ñ'=>'n','ò'=>'o','ó'=>'o','ô'=>'o','õ'=>'o','ù'=>'u','ú'=>'u','û'=>'u','ý'=>'y','ÿ'=>'y','Ä'=>'Ae',
        'ä'=>'ae','Ö'=>'Oe','Ø'=>'Oe','ö'=>'oe','ø'=>'oe','Ü'=>'Ue','ü'=>'ue','Þ'=>'TH','þ'=>'th','Ð'=>'DH','ð'=>'dh','ß'=>'ss','Œ'=>'OE','œ'=>'oe','Æ'=>'AE','æ'=>'ae','µ'=>'u'
    ));
    $s=strtolower($s);
    $s=preg_replace('@[^a-z0-9.-]@','.',$s);
    $s=preg_replace('@([.-])[.-]+@','$1',$s);
    $s=trim($s,'-.');
    if(!preg_match('@^[a-z0-9]([a-z0-9.-]*[a-z0-9])?$@',$s)) {
      $s='bad-name';
    }
    return $s;
  }

?>