#!/usr/bin/python2.7
#
# Univention DigiOnline WebWeaver integration
#
# Copyright (C) 2014 Univention GmbH <http://www.univention.de/>
# Copyright (C) 2014 DigiOnline GmbH <http://www.digionline.de/>
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import listener
import univention.debug as ud
import os
from time import time
import re
from errno import EEXIST
from pwd import getpwnam
# from json import dump

ID = '@%@safe_name@%@'
name = 'webweaver_' + ID
description = 'print all names/users/uidNumbers into a file'
filter = """(|
	(&(objectClass=posixAccount)(objectClass=shadowAccount))
	(objectClass=univentionMail)
	(objectClass=sambaSamAccount)
	(&(objectClass=person)(objectClass=organizationalPerson)(objectClass=inetOrgPerson))
	(objectClass=ucsschoolOrganizationalUnit)
	(univentionObjectType=groups/group)
)""".translate(None, '\t\n\r')
attributes = ['uid', 'uidNumber', 'cn', 'userPassword', 'objectClass', 'uniqueMember', 'description', 'mail']

USER_LIST = os.path.join('/var/spool/univention-digionline-webweaver', ID)
# FELDER = ('cn', 'o', 'univentionBirthday', 'title', 'userPassword', 'uid', 'Mail', 'displayName', 'mailPrimaryAddress', 'gecos', 'sn', 'givenName', 'departmentNumber', 'objectClass', 'univentionObjectType', 'gidNumber', 'description', 'memberUid', 'uniqueMember', 'entryUUID', 'entryDN')
SKIP_FIELDS = ('krb5Key', 'sambaHomePath', 'sambaProfilePath')
CONFIGURATION = '/etc/univention-digionline-webweaver.json'
BASE_DN = "ou=%s,%s" % (ID, listener.configRegistry['ldap/base'])


def handler(dn, new, old):
	if not dn.endswith(BASE_DN):
		return

	filename = os.path.join(USER_LIST, str(int(time())))
	with open(filename, 'a') as out:
		# dump({
		# 	'dn': dn,
		# 	'old': old,
		# 	'new': new,
		# }, out)
		print >> out, format_str(dn)
		print >> out, format_values(new)
		print >> out, format_values(old)


def initialize():
	""" Initialisierungsfunktion, die Output-Datei wird
	geloescht, falls vorhanden. """
	try:
		os.mkdir(USER_LIST, 0700)
		ud.debug(ud.LISTENER, ud.INFO, 'Directory "%s" created.' % (USER_LIST,))
	except OSError as ex:
		if ex.errno != EEXIST:
			ud.debug(ud.LISTENER, ud.INFO, 'Failed to create directory "%s": %s' % (USER_LIST, ex))
			raise
	ent = getpwnam('listener')
	os.chown(USER_LIST, ent.pw_uid, ent.pw_gid)
	for filename in os.listdir(USER_LIST):
		pathname = os.path.join(USER_LIST, filename)
		try:
			os.remove(pathname)
		except OSError as ex:
			ud.debug(ud.LISTENER, ud.WARN, 'Failed to delete file "%s": %s' % (pathname, ex))
#	filename = os.path.join(USER_LIST, '0')
#	with open(filename, 'a') as out:
#		print >> out, 'INIT'

def postrun():
	listener.run('/usr/bin/php5', [
		'php5',
		'/usr/share/univention-digionline-webweaver/univention-digionline-webweaver.php',
		CONFIGURATION,
		ID,
	])


def filter_values(values):
	r"""
	Print LDAP values.

	>>> from json import dumps
	>>> dumps(filter_values({'name': ['value1', 'value2']}))
	'{"name": ["value1", "value2"]}'
	>>> dumps(filter_values({'name': []}))
	'{"name": []}'
	"""
	return dict(
		(key, str(values[0]) if type(values) in (list, tuple) and len(values) == 1  else '' if len(values) == 0 else values)
		for (key, values) in values.iteritems()
		if name not in SKIP_FIELDS
	)


def format_values(values):
	r"""
	Print LDAP values.
	>>> format_values({'name': ['value1', 'value2']})
	'{"name": ["value1", "value2"]}'
	>>> format_values({'name': []})
	'{"name": }'
	"""
	return "{%s}" % (
		", ".join(
			"%s: %s" % (
				format_str(name),
				format_str(str(values[0])) if type(values) in (list, tuple) and len(values) == 1 \
				else '' if len(values) == 0 else "[%s]" % ", ".join(format_str(value) for value in values)
			)
			for (name, values) in values.items()
			if name not in SKIP_FIELDS
		),
	)


def format_str(string):
	r"""
	Escape double-quote and backslash characters.
	>>> format_str(r'foo')
	'"foo"'
	>>> format_str(r'abc"def')
	'"abc\\"def"'
	>>> format_str(r'abc\def')
	'"abc\\\\def"'
	"""
	return '"%s"' % (format_str.RE_ESCAPE.sub(lambda c: '\\' + c.group(0), string),)
format_str.RE_ESCAPE = re.compile(r'["\\]')


if __name__ == '__main__':
	import doctest
	print doctest.testmod()
